/*
	Definition av symboltabellen som lagrar symboler och vrden...
*/

#include <stdio.h>
#include <iostream.h>
#include <string.h>

typedef struct Tabellrad {	//Symboltabellen bestr av tabellrader
	char szNyckel[9];	 		//Nyckel
	int nVarde;    	 		//Vrde
   Tabellrad *pNasta; 		//Pekare till nsta tabellrad
} Tabellrad;

/*
	Klassen Hashtabell anvnds fr att hantera symboltabellen.
   Symboltabellen lagrar symboler (nycklar) och dess vrden.
	Hmtning och lagring i tabellen sker med hjlp av hashindexering,
   d.v.s. varje nyckel ger ett index till en bestmd rad i hashtabellen.
   Vid kollision (d.v.s. tv olika nycklar ger samma hashindex) lggs
	kolliderande nycklar i lnkade listor.
*/

class Hashtabell
{
	public:
   	Hashtabell(int nNyTabellStorlek);					 //Konstruktor
      ~Hashtabell();								          	 //Destruktor
      void SetNyckel(char szNyckel[9],int nNyttVarde); //Definierar nyckel
      int GetVarde(char szNyckel[9]);                	 //Hmtar nyckelns vrde
      int Hashing(char szNyckel[9]);                   //Ger index i tabellen
   private:
   	int nTabellStorlek;	                            //Storlek p hashtabell
      Tabellrad *pTabell;                          	 //Pekare till hashtabell
};

/*
	Konstruktorn till hashtabellen skapar och initierar ett frutbestmt antal
   tabellrader.
   Antal tabellrader bestms i filen findus.h. I detta fall har tabellraderna
   valts till 13 stycken.
*/

Hashtabell::Hashtabell(int nNyTabellStorlek)
{
    //Skapa en tabell som bestr av flera tabellrader:
    pTabell = new Tabellrad[nNyTabellStorlek];
    //Nollstll allt i tabellen:
    for (int i=0;i<nNyTabellStorlek;i++)
    {
        pTabell[i].szNyckel[0]=' ';		//r frsta tecknet " " r nyckeln "tom"
        pTabell[i].nVarde=0;
        pTabell[i].pNasta=NULL;  		//NULL = Vi lnkar ej vidare
    }
    nTabellStorlek=nNyTabellStorlek;	//Kom ihg tabellens storlek
}

/*
	Destruktorn till hashtabellen stdar upp efter sig. Den ser till att
   de dynamiskt allokerade tabellraderna avallokeras.
   Till sist avallokeras den statiskt allokerade tabellen, d.v.s.
   huvudtabellen.
*/

Hashtabell::~Hashtabell()
{
	for (int i=0;i<13;i++)	//Fr varje tabellrad...
   {
  		Tabellrad *temp;
      temp = &pTabell[i];	//Peka p frsta tabellraden, horizontellt sett
   	while (temp->pNasta!=NULL)	//S lnge som tabellraden "lnkar vidare"...
      {
	      temp = &pTabell[i];	//Peka p frsta tabellraden, horizontellt sett
			//Stega fram tills nst sista tabellrad, horizontellt sett:
	   	while ( (temp->pNasta)->pNasta!=NULL )
	      {
				temp=temp->pNasta;
	      }
	      delete temp->pNasta;	//Avallokera sista tabellraden
         temp->pNasta=NULL;  //Den sista tabellraden pekar inte vidare
      }
   }
	delete[] pTabell; //Avallokera tabellen
}

/*
	Hashing-funktionen anvnds fr att generera ett hashindex, som baseras
   p nyckelns namn och hasttabellens storlek.
   Hashindexet anvnds fr att peka ut rtt rad i hashtabellen.
   Detta ger en effektiv skning och lagring i hashtabellen, frutsatt
   att det inte sker mnga "kollisioner".
*/

int Hashtabell::Hashing(char szNyckel[9])    //Returnerar hashvrde (index)
{
    int nSumma = 0;
    for (int i=0;szNyckel[i]!='\0';i++)  //Algoritm som ger hashvrde
    {
        nSumma = (nSumma * 128 + szNyckel[i]) % nTabellStorlek;
    }
    return (nSumma); //Summan r mellan 0 och tabellens storlek-1
}

/*
	SetNyckel lagrar nyckel och dess vrde i hashtabellen.
   Lmplig rad fs genom att generera hashindex fr nyckeln.
   Om platsen i "utpekad" tabellrad r tom lagras nyckeln dr.
   Om platsen ej r tom, sker man tills man funnit slutet av
   eventuell lnkad lista. Nr man funnit slutet av lnkade listan
   s skapas en ny tabellrad, som lnkas in i den lnkade listan.
   Till slut lagras nyckeln och dess vrde p den nya raden.
*/

void Hashtabell::SetNyckel(char szNyckel[9],int nNyttVarde)
{
	int position = Hashing(szNyckel); //Generera index (hashvrde)

   if (pTabell[position].szNyckel[0]==' ')	//Om nyckelkolumnen r tom...
   {
   	//Kopiera in nyckel i hashtabellen
      strncpy(pTabell[position].szNyckel, szNyckel, 9);
      //Kopiera in vrdet i hashtabellen
      pTabell[position].nVarde = nNyttVarde;
   }
   else													//Om nyckelkolumn r upptagen...
   {
   	//Om Samma nyckel...
   	if (strcmp(pTabell[position].szNyckel,szNyckel)==0)
      {
      	pTabell[position].nVarde=nNyttVarde;    	//ndra vrdet p nyckeln
      }
      //Annars "Kollision"... (d.v.s. 2 nycklar har samma hashvrde)
      else
      {
      	Tabellrad *temp;
         temp = &pTabell[position];

         //Stega fram tills vi hittat NULL (listslut) eller samma nyckel
         while ((temp->pNasta!=NULL)&&(strcmp(temp->szNyckel,szNyckel)!=0))
         {
				temp=temp->pNasta;
         }

         if (strcmp(temp->szNyckel,szNyckel)!=0) //Om nyckeln ej finns....
         {
         	Tabellrad *pNyRad;
            pNyRad = new Tabellrad;					//Skapa ny tabellrad
            // Lgg in nya nyckeln och vrde i den nya raden och "NULLa" pNasta:
            strcpy(pNyRad->szNyckel, szNyckel);
            pNyRad->nVarde = nNyttVarde;
            pNyRad->pNasta = NULL;
            (temp->pNasta) = pNyRad; 	//Lnka frn tabell till ny rad
         }
         else												//Annars r nyckeln hittad...
         {
         	temp->nVarde=nNyttVarde;	 		//Uppdatera vrdet p nyckeln
         }
		}
   }
}

/*
	GetVarde hmtar nyckelns vrde frn hashtabellen. Skning sker med
   hjlp av hashindexering. Om rtt nyckel ej ptrffas i frsta frsket,
   letar man vidare i eventuell lnkad lista, tills man hittat nyckeln.
   Om nyckel ej hittas, anropas SetNyckel fr att lagra den nya nyckeln,
   som initieras till noll.
*/

int Hashtabell::GetVarde(char szNyckel[9])    	//Hmtar nyckelns vrde
{
    int position=Hashing(szNyckel);					//Hmta hashvrde (index)

    if (strcmp(pTabell[position].szNyckel, szNyckel)==0) //Om nyckel funnen...
    {
    	return pTabell[position].nVarde;  //Returnera nyckelvrde
    }
    else						  			//Annars leta vidare i den lnkade listan...
    {
    	Tabellrad *temp;
      temp = &pTabell[position];

    	//Stega fram tills vi hittat NULL (listslut) eller samma nyckel
      while ( ((temp->pNasta)!=NULL) && (strcmp(temp->szNyckel,szNyckel)!=0) )
      {
         temp=temp->pNasta;
      }

		//Om nyckeln ej finns....
      if( strcmp(temp->szNyckel,szNyckel)!=0 )
      {
      	SetNyckel(szNyckel,0);	//Skapa nyckel
         return 0;
      }
      //Annars r nyckeln hittad...
      else
      {
            return temp->nVarde;
      }
	}
}

